#!/usr/bin/env python
# -*- coding: utf-8 -*-

#Fichero:  gestionVlanSNMP.py
#Fecha última modificación: 31 de Octubre de 2019
#Autor: Darío Adánez Arroyo
#                                                                                                       
#Descripción:############################################################################################
#Este fichero se encarga de realizar los cambios necesarios en la configuración del switch cuando se    #
#ha recibido una trampa informado de la bajada o subida de un puerto o cuando un usuario se ha          #
#autenticado con éxito, así como de realizar los cambios pertinentes en la base de datos. Se vale de    #
#las herramientas proporcionadas por pysnmp y SQLite3 para realizar los cambios en los switches y en la #
#base de datos.                                                                                         #
#########################################################################################################

#Realiza las importaciones de paquetes necesarias
import sys
import signal
import os
from subprocess import call
import sqlite3
from pysnmp.proto.rfc1902 import *
from pysnmp.hlapi import *
from pysnmp import smi
from pysnmp.entity import engine, config
from pysnmp.carrier.asyncore.dgram import udp
from pysnmp.carrier import error
from pyasn1.codec.ber import decoder
from pyasn1.type import univ
import time

class gestionVlanSNMP:
    def __init__(self):
        #Realiza la conexión a la base de datos para consultar la configuración 
        con = sqlite3.connect("/home/dario/TFG/basedatos.db")
        cursorObj = con.cursor()
        consulta = "SELECT * FROM configuracion"
        
        cursorObj.execute(consulta)
        resultado = cursorObj.fetchone()
        
        #Obtiene los parametros de configuración
        self.ip = resultado[0]
        self.num_puertos = resultado[1]
        self.pid = resultado[2]
        self.vlan_restringida = resultado[3]
        self.vlan_registrados = resultado[4]
        self.vlan_conocidos = resultado[5]
        self.comunidad_lectura = resultado[6]
        self.comunidad_escritura = resultado[7]
        con.close()
    #########################################################################
    #                    Método snmpwalk                                    #
    #########################################################################
    #Se trata de un método auxiliar que se utiliza para realizar las        #
    #consultas iterativas de un OID. Recibe como argumentos el OID sobre el #
    #se va a realizar la consultas y al host al que debe dirigirse. Devuelve#
    #dos arrays, uno con el nombre del elemento y otro con el valor asignado#
    #########################################################################
    def snmpwalk(self,oid):
        
        #Arrays para devolver el resultado
        contenido = []
        valor = []
        #Realiza la acción
        try:
            
            for (errorIndication,
            errorStatus,
            errorIndex,
            varBinds) in nextCmd(SnmpEngine(),
            CommunityData(self.comunidad_lectura),
            UdpTransportTarget((self.ip, 161)),
            ContextData(),
            ObjectType(ObjectIdentity(oid)),
            lexicographicMode=False):
            
                if errorIndication:

                    print(errorIndication)

                    break

                elif errorStatus:

                    print("%s at %s" % (errorStatus.prettyPrint(),
                    errorIndex and varBinds[int(errorIndex) - 1][0] or "?"))

                    break

                else:
                    

                    for name, val in varBinds:
                        #Almacena el resultado en las variables
                        valor.append(val)
                        contenido.append(name)
        
        except:
            #Si ha ocurrido algún error muestra un mensaje que avisa de ello
            print "Ha ocurrido un error al realizar la acción"
            return
        return contenido, valor


    #########################################################################
    #                    Método snmpget                                     #
    #########################################################################
    #Se trata de un método auxiliar que se utiliza para realizar una        #
    #consultas sobre  un OID. Recibe como argumentos el OID sobre el  que se#
    #va a realizar la consulta y al host al que debe dirigirse. Devuelve el #
    #valor del elemento consultado.                                         #
    #########################################################################
    def snmpget(self,oid):
        #Realiza la acción
        try :
            
            g = getCmd(SnmpEngine(),
                CommunityData(self.comunidad_lectura),
                UdpTransportTarget((self.ip, 161)),
                ContextData(),
                ObjectType(ObjectIdentity(oid)))
            errorIndication, errorStatus, errorIndex , varBinds = next( g )
            
        except:
            #Si ha ocurrido algún error informa de ello.
            print "Ha ocorrido un error al realizar el get"
            return
        
        if errorIndication:
            print errorIndication
        elif errorStatus:
            print("%s at %s" % (errorStatus.prettyPrint(),
            errorIndex and varBinds[int(errorIndex) - 1][0] or "?"))
        else:
            for name, val in varBinds:
                #Devuelve el valor.
                return val

    #########################################################################
    #                    Método snmpset                                     #
    #########################################################################
    #Se trata de un método auxiliar utilizado para realizar una operación   #
    #SET y cambiar el estado de alguna variable en el switch. Recibe como   #
    #parámetros el OID del parámetro a modificar y el host sobre el que se  #
    #debe realizar la modificación.                                         #
    #########################################################################
    def snmpset(self,oid,valor):
        try:
            g = setCmd(SnmpEngine(),
                CommunityData(self.comunidad_escritura),
                UdpTransportTarget((self.ip, 161)),
                ContextData(),
                ObjectType(ObjectIdentity(oid), valor))
                    
            errorIndication, errorStatus, errorIndex , varBinds = next(g)
        except:
            print "Ha ocorrido un error al realizar el get"
            return
        if errorIndication:

                print(errorIndication)
                return

        elif errorStatus:

                print("%s at %s" % (errorStatus.prettyPrint(),
                errorIndex and varBinds[int(errorIndex) - 1][0] or "?"))
                return

        else:
            #No devuelve nada ya que no nos interesa el resultado de la operación set
            for name, val in varBinds:
                return

    #########################################################################
    #                       Método obtieneMAC                               #
    #########################################################################
    # Método que se encarga de obtener la dirección MAC del dispositivo que #
    # se acaba de conectar al puerto y ha generado la trampa. Recibe como   #
    # argumento el número de puerto que ha sufrido la conexión y devuelve la#
    # dirección MAC del dispositivo conectado.                              #
    #########################################################################
    def obtieneMAC(self,puerto):
        #Hace la consulta SNMP para obtener la relación de puerto-mac que hay en el switch
        respuesta,valor = self.snmpwalk("1.3.6.1.2.1.17.7.1.2.2.1.2")
        lista = []
        for i in range(0,len(valor)):
            #Si el resultado que se está evaluando es el del puerto que queremos obtener la MAC entra
            if(valor[i]==int(puerto)):
                
                #Separa la parte de la MAC de la parte del OID
                auxiliar  = str(respuesta[i]).split("1.3.6.1.2.1.17.7.1.2.2.1.2.")
                decimal = auxiliar[1].split(".")
                #Empieza en 1 porque el primer elemento corresponde a la vlan del puerto
                for i in range (1,len(decimal)):
                        #Guarda en una lista cada parte de la dirección MAC
                        #La función hex() devuelve el resultado de la forma "0x"
                        #Con la función replace lo eliminamos
                        if int(decimal[i]) > 15:
                            
                            lista.append(hex(int(decimal[i])).replace("0x","")) 
                        #Si el valor es menor de 15 (F), coloca un 0 delante para que cada octeto cuente con 2 dígitos
                        else:
                        
                            lista.append("0" + hex(int(decimal[i])).replace("0x","")) 
                return ":".join(lista)
        
        return None
    #########################################################################
    #                           Método link_down                            #
    #########################################################################
    # Método que se encarga de gestionar la desconexión de un puerto. Este  #
    # método es llamado cuando se produce la desxonexión de un puerto del   #
    # switch. Debe comprobar si el puerto es un puerto "protegido", además  #
    # de cambiar la entrada de la tabla switch correspondiente al puerto y  #
    # volver a configurar el puerto para que esté en la vlan restringida.   #
    #########################################################################
    def link_down(self,puerto):
        
        
        try:
            #Realiza la conexión a la base de datos
            con = sqlite3.connect("/home/dario/TFG/basedatos.db")
            cursorObj = con.cursor()
            #Si se trata de un puerto protegido solo actualiza el estado y la mac
            if (int(puerto) == 1 or int(puerto) == 2):
                #Sentencia sql
                sql_update = "UPDATE switch set estado = 2, mac = '" + str(puerto) + "' where puerto = " + puerto
                #Ejecuta la sentencia
                cursorObj.execute(sql_update)
                #Guarda los datos
                con.commit()
                con.close()
                try:
                    os.kill(self.pid,signal.SIGUSR1)
                except:
                    print "Ha ocurrido un error al enviar la señal proceso"
                return
            #Si se trata de un puerto "normal", comprueba si la desconexión la ha realiza la aplicación
            sql_consulta = "SELECT puerto_manipulado FROM switch where puerto = " + puerto
            cursorObj.execute(sql_consulta)
            resultado = cursorObj.fetchone()
        except:
            print "Ha ocurrido un error al obtener el campo auxiliar del puerto %s" % puerto
            return
        
         #Si el puerto se encontraba en una vlan distinta, deberá cambiar su estado
        sql_consulta = "SELECT vlan FROM switch where puerto = " + puerto
        cursorObj.execute(sql_consulta)
        vlan = cursorObj.fetchone()
        
        if resultado[0] == 1:
            con.close()
            #Significará que la desconexión del puerto la hemos provocado nosotros
            return #Lo deja todo como está
        else:
            try:
            
                sql_update = "UPDATE switch SET estado = 2, mac = null, vlan =" + str(self.vlan_restringida) +", login=null WHERE puerto = " + str(puerto)
                cursorObj.execute(sql_update)
                con.commit()
               
            except:
                
                print "Ha ocurrido un error al actualizar el estado del puerto %s" % puerto
                return
           
           
            if vlan[0] != self.vlan_restringida:
                
                self.cambia_vlan(puerto,vlan[0],self.vlan_restringida)
                
            con.close()
            try:
                os.kill(self.pid,signal.SIGUSR1)
            except:
                print "Ha ocurrido un error al enviar la señal proceso"
    #########################################################################
    #                           Método link_up                              #
    #########################################################################
    # Método que se encarga de gestionar la conexión de un puerto. Este     #
    # método es llamado cuando se recibe una trampa indicando la conexión de#
    # un puerto en el switch. Debe comprobar si el puerto es un puerto      #
    # "protegido", ademásde cambiar la entrada de la tabla switch           #
    # correspondiente al puerto y comprobar si es una mac conocida para     #
    # el puerto de VLAN.                                                    #
    #########################################################################
    def link_up(self,puerto):
        
        
        #Espera un segundo
        time.sleep(1)
        #Lo utilizaremos como contador de reintentos
        i = 0
        while True:
            
            if i == 10:
                return
            #Obtiene la dirección MAC del dispositivo conectado al puerto
            mac = self.obtieneMAC(puerto)
            #Si no recibe nada como respuesta, vuelve a intentarlo
            if mac is None:
                time.sleep(1)
            else:
                break
            i = i + 1
        try:
        
            #Realiza la conexión a la base de datos
            con = sqlite3.connect("/home/dario/TFG/basedatos.db")
            cursorObj = con.cursor()
            
            #Estos puertos están reservados para el router y para el servidor
            #asique no deben ser modificados en la base de datos
            if (int(puerto) == 1 or int(puerto) == 2):
                sql_update = "UPDATE switch set estado = 1, mac = '" + mac + "' where puerto = " + puerto
                cursorObj.execute(sql_update)
                con.commit()
                con.close()
                try:
                    os.kill(self.pid,signal.SIGUSR1)
                except:
                    print "Ha ocurrido un error al enviar la señal proceso"
                return
            #Realiza una consulta para comprobar si realmente se ha conectado un dispositivo nuevo
            sql_consulta = "SELECT puerto_manipulado FROM switch WHERE puerto = " + puerto 
            cursorObj.execute(sql_consulta)
            rows = cursorObj.fetchall()
        except: 
            print "Ha ocurrido un error al trabajar con la base de datos"
            return
        
        
        #Evalúa el resultado de la consulta
        for row in rows :
            #Significará que la activación del puerto se debe a un proceso de la aplicación y no a 
            #que se ha conectado un nuevo dispositivo 
            if row[0] == 1:
                try:
                    
                    #Vuelve a poner el campo auxiliar a 0 para que cuando se desconecte todo funcione con normalidad
                    sql_update = "UPDATE switch set puerto_manipulado = 0 where puerto = " + puerto
                    cursorObj.execute(sql_update)
                    con.commit()
                    con.close()
                    try:
                        os.kill(self.pid,signal.SIGUSR1)
                    except:
                        print "Ha ocurrido un error al enviar la señal proceso"
                    return
                except:
                    print "Ha ocurrido un error al obtener el campo auxiliar"
                    return
        try:
            
            #Realiza una consulta para ver si el dispositivo conectado es conocido
            sql_consulta = "SELECT * FROM pc_conocidos WHERE mac = '" + str(mac) + "'"
            cursorObj.execute(sql_consulta)
            resultado = cursorObj.fetchone()
            
            #Si no hay resultado significa que el dispositivo conectado no se encuentra dentro de los dispositivos conocidos
            if resultado is None: 
                #Realiza la actualización de la tabla para mostrar los cambios
                
                sql_update = "UPDATE switch SET estado = 1, mac = '" + str(mac) + "' , vlan =" + str(self.vlan_restringida) + " where puerto = " + puerto
                cursorObj.execute(sql_update)
                con.commit()

            else:
                
                #Realiza la actualización de la tabla para mostrar los cambios
                sql_update = "UPDATE switch SET puerto_manipulado = 1, estado = 1, mac = '" + str(mac) + "' , vlan = " + str(self.vlan_conocidos) + ", login ='" + resultado[1] + "' where puerto = " + puerto
                cursorObj.execute(sql_update)
                con.commit()

                #En este caso, además habría que cambiar al puerto de vlan
                self.cambia_vlan(puerto,self.vlan_restringida,self.vlan_conocidos)
                
                oid = "1.3.6.1.2.1.2.2.1.7." + puerto
                #Desactiva el puerto para que el host "olvide" su IP
                self.snmpset(oid,Integer32(2))
                #Introduce un retardo ya que si se vuelve a levantar el puerto al instante 
                #no se realiza una nueva petición DHCP
                time.sleep(5)
                #Levanta el puerto para que obtenga la IP y vuelva a tener conexión
                self.snmpset(oid,Integer32(1))
            con.close()
        except:
            print "Ha ocurrido un error al trabajar con la base de datos"
            return
        try:
            os.kill(self.pid,signal.SIGUSR1)
        except:
            print "El proceso ya no existe"
    #########################################################################
    #                           Método cambia_vlan                          #
    #########################################################################
    # Método que se encarga de cambiar un puerto de VLAN. Recibe como       #
    # parámetro el puerto, la vlan en la que se encuentra y a la que lo     #
    # queremos cambiar. No devuelve nada.                                   #
    #########################################################################
    def cambia_vlan(self,puerto,vlan,vlan_nueva):
        
        #OID para consultar el estado de la vlan en los puertos
        oid = "1.3.6.1.2.1.17.7.1.4.3.1.2." + str(vlan)
        #Obtiene el estado
        val = self.snmpget(oid)
        #Se queda solo con la parte importante (los primeros 8 bits que representan los puertos
        #nuestro switch
        #Obtiene el número de bits de la respuesta para poder operar
        num_bits = len(val) * 8
        #Concatena el resultado para trabajar con el
        r = val[0]
        for i in range(1,len(val)):
            r = r + val[i]

        #Convierte el tipo OctecString a entero para poder operar con el
        s = int(r.encode('hex'), 16)
        #En función del puerto que vaya a cambiar de vlan se realiza la resta para eliminar
        #la vlan del puerto en la variable dot1qVlanStaticEgressPorts
        s = s & ~(1<<(num_bits-int(puerto)))
        #Convierte a cadena hexadecial el entero y lo completa con 8 0s para que la operación no falle
        nuevo_valor = hex(s)
        #Cambia la variable dot1qVlanStaticEgressPorts de la vlan para eliminar la vlan del puerto
        self.snmpset(oid,OctetString(hexValue=nuevo_valor.replace("0x","")))
        
        # Ahora hay que hacer los mismos cambios pero para la nueva vlan
        
        oid = "1.3.6.1.2.1.17.7.1.4.3.1.2." + str(vlan_nueva)
        #Obtiene el estado
        val = self.snmpget(oid)
        #Se queda solo con la parte importante (los primeros 8 bits que representan los puertos
        #nuestro switch
        #Concatena el resultado para trabajar con el
        r = val[0]
        for i in range(1,len(val)):
            r = r + val[i]
        #Convierte el tipo OctecString a entero para poder operar con el
        s = int(r.encode('hex'), 16)
        #Añade el bit correspondiente al nuevo puerto
        s = s | (1<<(num_bits-int(puerto)))
        nuevo_valor = hex(s) 
        #Cambia la variable dot1qVlanStaticEgressPorts de la vlan para eliminar la vlan del puerto
        self.snmpset(oid,OctetString(hexValue=nuevo_valor.replace("0x","")))
        
        #Modifica el oid para este caso
        oid = "1.3.6.1.2.1.17.7.1.4.5.1.1." + str(puerto)
        #Finalmente hay que cambiar al puerto de vlan modificando  dot1qPvid
        self.snmpset(oid,Gauge32(vlan_nueva))
    #########################################################################
    #                      Método usuario_autenticado                       #
    #########################################################################
    # Método que es llamado cuando un usuario accede satisfactoriamente al  #
    # sistema mediante el formulario web. Recibe como parámetro el puerto   #
    # sobre el que trabajar. Se encarga de cambiar al puerto de vlan y      #
    # de forzar la conexión y desconexión del puerto para que obtenga la IP #
    # de la nueva red.                                                      #
    #########################################################################
    def usuario_autenticado(self,puerto,login):
    
        
        try:
            sql_update="UPDATE switch SET vlan=" + str(self.vlan_registrados) +", puerto_manipulado=1, login='" + login +"' where puerto = " + puerto
            #Realiza la conexión a la base de datos
            con = sqlite3.connect("/home/dario/TFG/basedatos.db")
            cursorObj = con.cursor()
            cursorObj.execute(sql_update)
            con.commit()
        except:
            print "No se ha podido ejecutar la consulta"
        
        con.close()
        self.cambia_vlan(puerto,self.vlan_restringida,self.vlan_registrados);
        oid = "1.3.6.1.2.1.2.2.1.7." + str(puerto)
        #Desactiva el puerto para que el host "olvide" su IP
        self.snmpset(oid,Integer32(2))
        #Introduce un retardo ya que si se vuelve a levantar el puerto al instante 
        #no se realiza una nueva petición DHCP
        time.sleep(10)
        #Levanta el puerto para que obtenga la IP y vuelva a tener conexión
        self.snmpset(oid,Integer32(1))
    def importa_datos(self):
        
        con = sqlite3.connect("/home/dario/TFG/basedatos.db")
        cursorObj = con.cursor()
        #1.3.6.1.2.1.2.2.1.8
        cambio = 0
        cursorObj.execute("SELECT * FROM switch")
        base = cursorObj.fetchall()
        oid,estado = self.snmpwalk("1.3.6.1.2.1.2.2.1.8")
        for i in range(0,int(self.num_puertos)):
            print i
            if estado[i] != base[i][1]:
                cambio = 1
                sql_update ="UPDATE switch SET estado =" + str(estado[i]) + " WHERE puerto =" + str(i+1)
                cursorObj.execute(sql_update)
                con.commit()
        for i in range(0,int(self.num_puertos)):
            mac= self.obtieneMAC(i+1)
            if mac != base[i][3]:
                if mac: 
                    
                    cambio = 1
                    sql_update= "UPDATE switch SET mac ='" + mac + "' WHERE puerto=" + str(i+1)
                    cursorObj.execute(sql_update)
                    con.commit()
                else:
                    cambio = 1
                    sql_update= "UPDATE switch SET mac = null WHERE puerto=" + str(i+1)
                    cursorObj.execute(sql_update)
                    con.commit()
        oid1,vlan = self.snmpwalk("1.3.6.1.2.1.17.7.1.4.5.1.1")
        for i in range(0,int(self.num_puertos)):
            if vlan[i] != base[i][5]:
                cambio = 1
                sql_update= "UPDATE switch SET vlan=" + str(vlan[i]) + " WHERE puerto=" + str(i+1)
                cursorObj.execute(sql_update)
                con.commit()
        con.close()
        return cambio 
if __name__ == "__main__":
    #Realiza la conexión a la base de datos para consultar la configuración 
    obj = gestionVlanSNMP()
   
    #Comprueba cual es el primer algumento de llamada para saber que tipo de evento se ha producido
    if (str(sys.argv[1]) == "down"):
        obj.link_down(sys.argv[2])
        
    elif(str(sys.argv[1]) == "up"):
        obj.link_up(sys.argv[2])
        
    elif(str(sys.argv[1]) == "login"):

        obj.usuario_autenticado(sys.argv[2],str(sys.argv[3]))
        
 
